/* /////////////////////////////////////////////////////////////////////////////
 * File:        iwhereis.cpp
 *
 * Purpose:     Adaptation of the Synesis Software whereis utility for Internet
 *              (FTP) enumeration
 *
 * Created:     15th May 2004
 * Updated:     21st May 2005
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 *              The timestr() function and the verbose formatting in trace_file
 *              used by kind permission of Walter Bright and Digital Mars.
 *
 * License:     (Licensed under the Synesis Software Standard Public License)
 *
 *              Copyright (C) 1999-2005, Synesis Software Pty Ltd.
 *              Copyright (C) 1987-2004, Digital Mars.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au
 *                          http://www.synesis.com.au/software
 *
 *                          http://www.digitalmars.com/
 *
 *              email:      software@synesis.com.au
 *
 *                          contact@digitalmars.com
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Redistributions in binary form must reproduce the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer in the documentation
 *              and/or other materials provided with the distribution.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iv) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (v) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <stdio.h>

#include <stlsoft.h>

stlsoft_ns_using(ss_sint64_t)
stlsoft_ns_using(ss_bool_t)

/* Regrettably, something in this program is too complex for early versions of
 * many compilers that the STLSoft libraries support. With some it is a compiler
 * issue, with others it either crashes or produces erroneous output. Thus, the
 * following restrictions are mandated:
 */
#if defined(__STLSOFT_COMPILER_IS_BORLAND)
# if __BORLANDC__ < 0x0560
#  error whereis.cpp cannot be successfully built by Borland C/C++ prior to version 5.6
# endif /* __BORLANDC__ < 0x0560 */
#elif defined(__STLSOFT_COMPILER_IS_DMC)
# if __DMC__ < 0x0832
#  error whereis.cpp cannot be successfully built by Digital Mars C/C++ prior to version 8.32
# endif /* __DMC__ < 0x0832 */
#elif defined(__STLSOFT_COMPILER_IS_GCC)
# if __GNUC__ < 3 && \
     __GNUC_MINOR__ < 95
#  error whereis.cpp cannot be successfully built by GNU C/C++ prior to version 6.0
# endif /* __GNUC__ < 3 && __GNUC_MINOR__ < 95 */
#elif defined(__STLSOFT_COMPILER_IS_INTEL)
# if __INTEL_COMPILER < 600
#  error whereis.cpp cannot be successfully built by Intel C/C++ prior to version 6.0
# endif /* __INTEL_COMPILER < 600 */
#elif defined(__STLSOFT_COMPILER_IS_MWERKS)
 /* No problems with Metrowerks */
#elif defined(__STLSOFT_COMPILER_IS_MSVC)
# if _MSC_VER < 1100
#  error whereis.cpp cannot be successfully built by Microsoft Visual C/C++ prior to version 5.0
# endif /* _MSC_VER < 1100 */
#else
# error Unrecognised compiler
#endif /* compiler */

/* /////////////////////////////////////////////////////////////////////////////
 * Warnings
 */

#if defined(__STLSOFT_COMPILER_IS_MSVC)
# pragma warning(disable : 4702)
#endif /* __STLSOFT_COMPILER_IS_INTEL */

/* ////////////////////////////////////////////////////////////////////////// */

#ifdef _SYNSOFT_INTERNAL_BUILD
# include "MLCompId.h"
#else /* ? _SYNSOFT_INTERNAL_BUILD */
# define _nameSynesisSoftware               "Synesis Software (Pty) Ltd"
# define _wwwSynesisSoftware_SystemTools    "http://synesis.com.au/systools.html"
# define _emailSynesisSoftware_SystemTools  "software@synesis.com.au"
#endif /* _SYNSOFT_INTERNAL_BUILD */

/* ////////////////////////////////////////////////////////////////////////// */

#include <stlsoft/auto_buffer.hpp>

stlsoft_ns_using(auto_buffer)

#include <stlsoft/limit_traits.h>

#include <stlsoft/malloc_allocator.hpp>

stlsoft_ns_using(malloc_allocator)

#include <stlsoft/string_access.hpp>

stlsoft_ns_using(c_str_ptr)

#include <stlsoft/simple_algorithms.hpp>

stlsoft_ns_using(remove_duplicates_from_unordered_sequence)

#include <stlsoft/string_tokeniser.hpp>

stlsoft_ns_using(string_tokeniser)

#if defined(UNIX) || \
    defined(unix) || \
    defined(__unix) || \
    defined(__unix__) || \
    (   defined(__xlC__) && \
        defined(_POWER) && \
        defined(_AIX))

# error There is currently no UNIX version of this program

#elif defined(WIN32) || \
      defined(_WIN32)

/* Win32 includes */
# define WHEREIS_PLATFORM_IS_WIN32
# define WHEREIS_PATH_MAX               (_MAX_PATH)

# include <inetstl.h>

# include <inetstl/findfile_sequence.hpp>

typedef inetstl::findfile_sequence_a    findfile_sequence;

# include <inetstl/functionals.hpp>

 using inetstl::compare_path;

# include <inetstl/filesystem_traits.hpp>

 using inetstl::filesystem_traits;

# include <winstl/error_desc.hpp>

 using winstl::error_desc;

# include <inetstl/session.hpp>

 using inetstl::session;

# include <inetstl/connection.hpp>

 using inetstl::connection;

#else

# error Operating system not recognised

#endif /* operating system */
#include <algorithm>

using std::copy;
using std::for_each;
using std::remove_if;

#include <functional>

using std::unary_function;

#include <iterator>

using std::back_inserter;

#include <string>

using std::string;

#include <vector>

using std::vector;

#include <list>

using std::list;

/* /////////////////////////////////////////////////////////////////////////////
 * Warnings
 */

#if defined(__STLSOFT_COMPILER_IS_INTEL)
# pragma warning(disable : 383)
# pragma warning(disable : 444)
# pragma warning(disable : 1419)
#endif /* __STLSOFT_COMPILER_IS_INTEL */

/* /////////////////////////////////////////////////////////////////////////////
 * Tracing
 */

//#define WHEREIS_TRACE

#if !defined(WHEREIS_TRACE) && \
    defined(_DEBUG)
# define WHEREIS_TRACE
#endif /* !WHEREIS_TRACE && _DEBUG */

/* /////////////////////////////////////////////////////////////////////////////
 * Version information
 *
 * Normally, Synesis Software components & tools are "versioned" by making use
 * of tool-generated header files, one of which - MBldHdr.h - defines symbolic
 * constants representing major, minor, revision and build numbers.
 *
 *
 * Since this file is to be available in a more public forum, including on the
 * Digital Mars site, these constants are merely defined here, rather than
 * encumber any developers who may wish to compile it with a lot of
 * Synesis-specific gunk.
 */

#ifdef _SYNSOFT_INTERNAL_BUILD
# include "MBldHdr.h"
# include "showver.h"

extern "C" const int  _mccVerHi   =   __SYV_MAJOR;
extern "C" const int  _mccVerLo   =   __SYV_MINOR;
extern "C" const int  _mccVerRev  =   __SYV_REVISION;
extern "C" const int  _mccBldNum  =   __SYV_BUILDNUMBER;

#else /* ? _SYNSOFT_INTERNAL_BUILD */
extern "C" const int  _mccVerHi   =   1;
extern "C" const int  _mccVerLo   =   0;
extern "C" const int  _mccVerRev  =   2;
extern "C" const int  _mccBldNum  =   3;
#endif /* _SYNSOFT_INTERNAL_BUILD */

# define	COMPANY_NAME	"Synesis Software"
# define	TOOL_NAME		"File Searching"
# define	EXE_NAME		"whereis"

/* /////////////////////////////////////////////////////////////////////////////
 * Typedefs
 */

typedef string_tokeniser<string, char>  tokeniser_string_t;
typedef vector<string>                  searchpath_sorted_t;
//typedef   list<string>                searchpath_sorted_t;

#if defined(WHEREIS_PLATFORM_IS_WIN32)
typedef FILETIME                        whereis_time_t;
#elif defined(WHEREIS_PLATFORM_IS_UNIX)
typedef time_t                          whereis_time_t;
#else
# error Unknown platform
#endif /* platform */

/* /////////////////////////////////////////////////////////////////////////////
 * Enumerations
 */

namespace Trim
{
	enum Trim
	{
			full
		,	fileOnly
#if 0
		// Supporting these two will require that deriverelativepath be able to
		// return the length required, given a null dest
		,	fromSearchRoot
		,	fromCurrentDirectory
#endif /* 0 */
	};
} // namespace Trim

/* /////////////////////////////////////////////////////////////////////////////
 * Macros and definitions
 *
 * Visual C++, sigh. uses %I64d", rather than "%lld", so that is abstracted here
 */

#if defined(__STLSOFT_COMPILER_IS_BORLAND) || \
    /* defined(__STLSOFT_COMPILER_IS_GCC) || */ \
    defined(__STLSOFT_COMPILER_IS_INTEL) || \
    defined(__STLSOFT_COMPILER_IS_MSVC)
#define _FMT_SINT64_WIDTH(x)            "%" #x "I64d"
#else
#define _FMT_SINT64_WIDTH(x)            "%" #x "lld"
#endif /* compiler */

/* /////////////////////////////////////////////////////////////////////////////
 * Constants
 */

#ifdef __STLSOFT_COMPILER_IS_INTEL
# pragma warning(disable : 1418)
#endif /* __STLSOFT_COMPILER_IS_INTEL */

/* Multi-part path/pattern delimiter */

#if defined(__STLSOFT_COMPILER_IS_MSVC) && \
	_MSC_VER < 1300
char const  PATH_SEPARATOR				=   ';';
#else /* ? compiler */
char const  PATH_SEPARATOR              =   filesystem_traits<char>::path_separator();
#endif /* compiler */

/* Path component delimiter */

#if defined(__STLSOFT_COMPILER_IS_MSVC) && \
	_MSC_VER < 1300
char const  PATH_NAME_SEPARATOR			=	'\\';
#else /* ? compiler */
char const  PATH_NAME_SEPARATOR			=	filesystem_traits<char>::path_name_separator();
#endif /* compiler */

/* Subdirectory wildcard pattern */

#if defined(__STLSOFT_COMPILER_IS_MSVC) && \
	_MSC_VER < 1300
char const  *PATTERN_ALL                =   "*";
#else /* ? compiler */
char const  *PATTERN_ALL                =   filesystem_traits<char>::pattern_all();
#endif /* compiler */

/* Illegal pattern characters string */

#if defined(_MSC_VER) /* For testing */ || \
    defined(WHEREIS_PLATFORM_IS_WIN32)
char const  *BAD_PATTERN_CHARS          =   ":\\";
#elif defined(WHEREIS_PLATFORM_IS_UNIX)
char const  *BAD_PATTERN_CHARS          =   "/";
#else
# error Unknown platform
#endif /* platform */

#ifdef __STLSOFT_COMPILER_IS_INTEL
# pragma warning(default : 1418)
#endif /* __STLSOFT_COMPILER_IS_INTEL */

/* /////////////////////////////////////////////////////////////////////////////
 * Function declarations
 *
 * (Primarily to keep Metrowerks happy, but that's a good thing to do!)
 */

static char *timestr(char *buffer, size_t cchBuffer, whereis_time_t const *ft); // Make Metrowerks happy

/* /////////////////////////////////////////////////////////////////////////////
 * Helper Functions
 */

char *timestr(char *buffer, size_t cchBuffer, whereis_time_t const *t)
{
#if defined(WHEREIS_PLATFORM_IS_WIN32)
    FILETIME    ftLocal;
    SYSTEMTIME  st;
    char        dateString[41];
    char        timeString[41];

    ::FileTimeToLocalFileTime(t, &ftLocal);
    ::FileTimeToSystemTime(&ftLocal, &st);

    ::GetDateFormat(LOCALE_USER_DEFAULT, 0, &st, NULL, dateString, stlsoft_num_elements(dateString));
    ::GetTimeFormat(LOCALE_USER_DEFAULT, 0, &st, NULL, timeString, stlsoft_num_elements(timeString));

#if 0
    // Digital Mars format (US based)
    sprintf(buffer, "%02d/%02d/%02d  %2d:%02d", st.wMonth, st.wDay, st.wYear % 100, st.wHour, st.wMinute);
#else /* ? 0 */
    // User-locale format
    sprintf(buffer, "%s  %s", dateString, timeString);
#endif /* 0 */

	STLSOFT_SUPPRESS_UNUSED(cchBuffer);

#elif defined(WHEREIS_PLATFORM_IS_UNIX)

    struct tm   tm_ = *localtime(t);

    strftime(buffer, cchBuffer, "%c", &tm_);

#else
# error Unknown platform
#endif /* platform */

    return buffer;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Functionals
 */

// struct trace_file
//
// This functional traces individual file records to stdout.

struct trace_file
    : public unary_function<findfile_sequence::value_type const &, void>
{
	typedef	trace_file	class_type;

public:
    trace_file(ss_bool_t bVerbose, char const *searchRoot, Trim::Trim trim)
        : m_bVerbose(bVerbose)
		, m_searchRoot(searchRoot)
		, m_trim(trim)
    {}

    void operator ()(findfile_sequence::value_type const &entry) const
    {
		size_t											len = ::stlsoft::c_str_len(entry);
		auto_buffer<char, malloc_allocator<char>, 256>	path(1 + len);

		switch(m_trim)
		{
			default:
				stlsoft_assert(0);
			case	Trim::full:
				strcpy(&path[0], c_str_ptr(entry));
				break;
			case	Trim::fileOnly:
				strcpy(&path[0], entry.get_filename());
				break;
#ifdef _SYNSOFT_INTERNAL_BUILD
#if 0
			case	Trim::fromSearchRoot:
				SynesisCrt::deriverelativepath(c_str_ptr(entry), m_searchRoot, path);
				break;
			case	Trim::fromCurrentDirectory:
				SynesisCrt::deriverelativepath(c_str_ptr(entry), s_szRoot, path);
				break;
#endif /* 0 */
#endif /* _SYNSOFT_INTERNAL_BUILD */
		}

        if(m_bVerbose)
        {
            ss_sint64_t                             size;
            char                                    timebuf[50];
            char                                    atts[8 + 1];
#if defined(WHEREIS_PLATFORM_IS_WIN32)
            findfile_sequence::find_data_type const &findData   =   entry.get_find_data();
            DWORD const                             &att        =   findData.dwFileAttributes;
            int                                     i;

            for (i = 0; i < 8; ++i)
            {
                static char tab[] = "RHSVDAXX";

                if (att & (1 << i))
                    atts[i] = tab[i];
                else
                    atts[i] = '-';
            }
            atts[8] = 0;

            timestr(timebuf, stlsoft_num_elements(timebuf), &findData.ftLastWriteTime);
            size = findData.nFileSizeHigh * __STLSOFT_GEN_SINT64_SUFFIX(0x100000000) + findData.nFileSizeLow;
#elif defined(WHEREIS_PLATFORM_IS_UNIX)
# ifdef _MSC_VER // For testing
            struct _stat    st;

            _stat(entry, &st);
# else /* ? _MSC_VER */
            struct stat     st;

            stat(entry, &st);
# endif /* _MSC_VER */

            atts[0] = ((st.st_mode & S_IWRITE) == 0) ? 'R' : '-';   // R
            atts[1] = '-'; // H
            atts[2] = '-'; // S
            atts[3] = '-'; // V
            atts[4] = ((st.st_mode & S_IFMT) == S_IFDIR) ? 'D' : '-';   // D
            atts[5] = '-'; // A
            atts[6] = '-'; // X
            atts[7] = '-'; // X
            atts[8] = 0;

            size = st.st_size;
            timestr(timebuf, stlsoft_num_elements(timebuf), &st.st_mtime);
#else /* ? platform */
# error Unknown platform
#endif /* platform */
            printf("%-22s  %s  " _FMT_SINT64_WIDTH(9) "\t%s\n", timebuf, atts, size, path.data());
        }
        else
        {
            fprintf(stdout, "%s\n", path.data());
        }
    }

// Members
private:
    ss_bool_t const		m_bVerbose;
	char const *const	m_searchRoot;
    Trim::Trim const	m_trim;

// Not to be implemented
private:
	class_type &operator =(class_type const &);
};

static void enumerate_dir_files(HINTERNET hconn, char const *dir, char const *pattern)
{
	typedef	findfile_sequence						sequence_t;

	sequence_t	fs(hconn, dir, pattern, PATH_SEPARATOR, findfile_sequence::files);

	sequence_t::const_iterator	begin	=	fs.begin();
	sequence_t::const_iterator	end		=	fs.end();

	std::for_each(begin, end, trace_file(true, "/", Trim::full));
#if 0
	for(; begin != end; ++begin)
	{
		printf("%s\n", (*begin).get_path());
	}
#endif /* 0 */
}

static void enumerate_dir(HINTERNET hconn, char const *dir, char const *pattern)
{
	enumerate_dir_files(hconn, dir, pattern);

	typedef	vector<string>	dirs_t;

	dirs_t	dirs;

	{
		dirs.reserve(32);

		SetLastError(ERROR_SUCCESS);

		findfile_sequence	fs(hconn, dir, "*", findfile_sequence::directories);

		findfile_sequence::const_iterator	begin	=	fs.begin();
		findfile_sequence::const_iterator	end		=	fs.end();
		DWORD								dwErr	=	::GetLastError();

		if(	ERROR_SUCCESS != dwErr &&
			ERROR_NO_MORE_FILES != dwErr)
		{
			printf("Failed to enumerate directories: %s (%lu)\n", stlsoft::get_ptr(winstl::error_desc(dwErr)), dwErr);
		}

		for(; begin != end; ++begin)
		{
			dirs.push_back((*begin).get_path());
		}
	}

	dirs_t::const_iterator	begin	=	dirs.begin();
	dirs_t::const_iterator	end		=	dirs.end();

	for(; begin != end; ++begin)
	{
		enumerate_dir(hconn, (*begin).c_str(), pattern);
	}

}

#if 0
// struct trace_dir
//
// This functional provides processing of a directory, used by the recursive
// option. It process the current directory (via process_path), and then all
// subdirectories (via trace_dir) "calling" itself recursively.

struct trace_dir
    : public unary_function<string const &, void>
{
	typedef	trace_dir	class_type;

public:
    trace_dir(string const &searchSpec, ss_bool_t bVerbose, Trim::Trim trim)
        : m_searchSpec(searchSpec)
        , m_bVerbose(bVerbose)
		, m_searchRoot(NULL)
		, m_trim(trim)
    {}
    trace_dir(string const &searchSpec, ss_bool_t bVerbose, char const *searchRoot, Trim::Trim trim)
        : m_searchSpec(searchSpec)
        , m_bVerbose(bVerbose)
		, m_searchRoot(searchRoot)
		, m_trim(trim)
    {}

    void operator ()(char const *dir) const
    {
		char			szSearchRoot[1 + MAX_PATH_LENGTH];
		char const		*searchRoot = (filesystem_traits<char>::get_full_path_name((NULL != m_searchRoot) ? m_searchRoot : dir, stlsoft_num_elements(szSearchRoot), szSearchRoot), szSearchRoot);
        process_path    f(m_searchSpec, m_bVerbose, searchRoot, m_trim);

        f(dir);

        findfile_sequence   directories(c_str_ptr(dir), PATTERN_ALL, findfile_sequence::directories);

        for_each(directories.begin(), directories.end(), trace_dir(m_searchSpec, m_bVerbose, searchRoot, m_trim));

		enumerate_dir(conn, /* rootDir */c_str_ptr(*sorted.begin()), searchSpec);
    }

    template <ss_typename_param_k S>
    void operator ()(S const &dir) const
    {
        operator ()(c_str_ptr(dir));
    }

private:
    string const		m_searchSpec;
    ss_bool_t const		m_bVerbose;
	char const *const	m_searchRoot;
    Trim::Trim const	m_trim;

// Not to be implemented
private:
	class_type &operator =(class_type const &);
};
#endif /* 0 */

/* /////////////////////////////////////////////////////////////////////////////
 * Function declarations
 */

static void usage(int bExit, char const *reason);
#ifdef WHEREIS_PLATFORM_IS_WIN32
static char const *get_driveroots(char *buffer);
#endif /* WHEREIS_PLATFORM_IS_WIN32 */

/* /////////////////////////////////////////////////////////////////////////////
 * Helper functionals
 */

struct is_empty_path
{
    template <ss_typename_param_k S>
    bool operator ()(S const &s) const
    {
        return c_str_ptr(s)[0] == '\0';
    }
};

/* /////////////////////////////////////////////////////////////////////////////
 * main
 */

#ifdef WIN32
extern "C" __declspec(dllimport) void __stdcall Sleep(unsigned long );
#endif /* WIN32 */

int main(int argc, char **argv)
{
    if(argc < 2)
	{
		usage(1, NULL);
	}
#if 0
	else if(argc > 5)
    {
        usage(1, "Invalid argument(s) specified");
    }
#endif /* 0 */
    else
    {
        bool        bSearchGivenRoots   =   false;
        bool        bSearchCwd          =   false;
        bool        bVerbose            =   true;
        bool        bRecursive          =   false;
        bool        bSuppressRecursive  =   false;
		char const	*host				=	NULL;
        char const  *rootSpec           =   NULL;
        char const  *searchSpec         =   NULL;
		char const	*username			=	NULL;
		char const	*password			=	NULL;
#ifdef WHEREIS_PLATFORM_IS_WIN32
        char        driveroots[105]     =   "";
#endif /* WHEREIS_PLATFORM_IS_WIN32 */

        for(int i = 1; i < argc; ++i)
        {
            char const  *arg    =   argv[i];

            if(arg[0] == '-')
            {
#ifdef _SYNSOFT_INTERNAL_BUILD
                if(0 == strcmp("--version", arg))
                {
                    return show_version();
                }
                else
#endif /* _SYNSOFT_INTERNAL_BUILD */
                {
					switch(arg[1])
					{
						case	'?':
							usage(1, NULL);
							break;
						case    'w':
						case    'W':
							bSearchCwd          =   true;
							break;
#ifdef WHEREIS_PLATFORM_IS_WIN32
						case    'h':
						case    'H':
							bSearchGivenRoots   =   true;
							rootSpec = get_driveroots(driveroots);
							break;
#endif /* WHEREIS_PLATFORM_IS_WIN32 */
						case    'r':
							bSearchGivenRoots   =   true;
							rootSpec = arg + 2;
							break;
						case    'n':
							username = arg + 2;
							break;
						case    'p':
							password = arg + 2;
							break;
						case    'R':
							bSuppressRecursive  =   true;
							break;
						case    'v':
						case    'V':
							bVerbose            =   true;
							break;
						case    'u':
						case    'U':
							bRecursive          =   true;
							break;
						case    's':
						case    'S':
							bVerbose            =   false;
							break;
						default:
	//                        fprintf(stderr, "Unrecognised argument \"%s\"\n", arg);
							usage(1, "Invalid argument(s) specified");
							break;
					}
				}
			}
			else
			{
				if(host == NULL)
				{
					host = arg;
				}
				else if(rootSpec == NULL)
				{
					rootSpec = arg;
				}
				else if(searchSpec == NULL)
				{
					searchSpec = arg;
				}
				else
				{
					usage(1, "<root-paths> and <search-spec> already specified");
				}
			}
		}

        if( ((bSearchCwd != false) +
            (bSearchGivenRoots != false)) > 1)
        {
            usage(1, "Cannot specify two or more of -w, -r, -e, -l, -i, -p");
        }

        // Validate the args. This entails just moving the root-dir into the
        // search-spec if no root-dir was specified. This simplifies the
        // previous command-line processing.
        if(NULL == searchSpec)
        {
            if( NULL != rootSpec &&
                !bSearchGivenRoots)
            {
                searchSpec = rootSpec;
                rootSpec = NULL;
            }
            else
            {
                usage(1, "Unexpected condition. Please report to Synesis Software");
            }
        }

        char    _rootDir[WHEREIS_PATH_MAX + 1];

        // If no directory is specified, the current directory is assumed
        if( rootSpec == NULL ||
            *rootSpec == '\0')
        {
            strcpy(_rootDir, "/");
            rootSpec = _rootDir;
        }

		if(NULL == host)
		{
			usage(1, "Must specify a host");
		}

		if(NULL == username)
		{
			username = "anonymous";
		}

        // Handle illegal characters
        if(NULL != strpbrk(searchSpec, BAD_PATTERN_CHARS))
        {
			char	sz[101];

            sprintf(sz, "<search-spec> contains illegal characters: \"%s\"\n", searchSpec);
            usage(1, sz);
        }

        searchpath_sorted_t     sorted;

        // Commence the search
        tokeniser_string_t      roots(rootSpec, PATH_SEPARATOR);

        copy(roots.begin(), roots.end(), back_inserter(sorted));

        // The roots may contain duplicate elements, so
        remove_duplicates_from_unordered_sequence(sorted, compare_path<char>());

        // They may also contain empty paths, so
        sorted.erase(remove_if(sorted.begin(), sorted.end(), is_empty_path()), sorted.end());

		for(;;)
		{
			try
			{
				/* . */
				session		sess;
				connection	conn(sess, host, INTERNET_INVALID_PORT_NUMBER, username, password, INTERNET_SERVICE_FTP, INTERNET_FLAG_PASSIVE, 0);

				// Now execute over the duplicate-free (but still in the correct search order)
				// sequence, applying the trace_dir functional to each one.
				searchpath_sorted_t::const_iterator	b	=	sorted.begin();
				searchpath_sorted_t::const_iterator	e	=	sorted.end();

				for(; b != e; ++b)
				{
					if(!bSuppressRecursive)
					{
						enumerate_dir(conn, /* rootDir */c_str_ptr(*b), searchSpec);
					}
					else
					{
						enumerate_dir_files(conn, /* rootDir */c_str_ptr(*b), searchSpec);
					}
				}

				break;
			}
			catch(inetstl::internet_exception x)
			{
				DWORD	err	=	x.error();

				if(ERROR_INTERNET_LOGIN_FAILURE == err)
				{
					if(NULL == password)
					{
						char	szUsername[101];
						char	szPassword[101];

						printf("Anonymous login not supported. Enter username and password details:\n");
						printf("Username: ");
						scanf("%100s", szUsername);
						username = szUsername;
						printf("Password: ");
						scanf("%100s", szPassword);
						password = szPassword;

						continue;
					}
				}

				printf("Exception caught: %s (%lu, %s)\n", x.what(), err, stlsoft::get_ptr(winstl::error_desc(err, "wininet")));

				break;
			}
		}

        STLSOFT_SUPPRESS_UNUSED(bVerbose);
        STLSOFT_SUPPRESS_UNUSED(bRecursive);
    }

    return 0;
}

/* /////////////////////////////////////////////////////////////////////////////
 * usage
 */

void usage(int bExit, char const *reason)
{
	FILE	*stm	=	(NULL == reason) ? stdout : stderr;

    fprintf(stm, "\n");
	fprintf(stm, "  %s %s Tool, v%d.%d.%d.%04d\n", COMPANY_NAME, TOOL_NAME, _mccVerHi, _mccVerLo, _mccVerRev, _mccBldNum);
    fprintf(stm, "\n");
    fprintf(stm, "    incorporating Digital Mars technology\n");
    fprintf(stm, "\n");

	if(NULL != reason)
	{
		fprintf(stm, "  Error: %s\n", reason);
		fprintf(stm, "\n");
	}

#ifdef _SYNSOFT_INTERNAL_BUILD
    fprintf(stm, "  USAGE 1: %s <host> [-n<username>] [-p<password>] [{-w | -r<root-paths> | -i | -l | -e<env-var> | -h}] [-d] [{-v | -s}] [-f | -F | | -t | -T] [<root-paths>] <search-spec>\n", EXE_NAME);
#else /* ? _SYNSOFT_INTERNAL_BUILD */
    fprintf(stm, "  USAGE 1: %s <host> [-n<username>] [-p<password>] [{-w | -r<root-paths> | -i | -l | -e<env-var> | -h}] [-d] [{-v | -s}] [-f | -F] [<root-paths>] <search-spec>\n", EXE_NAME);
#endif /* _SYNSOFT_INTERNAL_BUILD */
    fprintf(stm, "    where:\n\n");
    fprintf(stm, "    -n<username>   - specifies the username with which to log onto the host;\n");
	fprintf(stm, "                     defaults to \"anonymous\"\n");
    fprintf(stm, "    -p<password>   - specifies the password to use\n");
    fprintf(stm, "    -d             - displays the path(s) searched\n");
    fprintf(stm, "    -f             - shows the filename and extension only\n");
    fprintf(stm, "    -F             - shows the full path; (default) \n");
    fprintf(stm, "    -h             - searches from the roots of all drives on the system\n");
#if defined(WHEREIS_PLATFORM_IS_WIN32)
    fprintf(stm, "    -r<root-paths> - searches from the given root path(s), separated by \';\',\n");
    fprintf(stm, "                     eg.\n");
    fprintf(stm, "                       -r\"c:\\windows;x:\\bin\"\n");
#elif defined(WHEREIS_PLATFORM_IS_UNIX)
    fprintf(stm, "    -r<root-paths> - searches from the given root path(s), separated by \':\',\n");
    fprintf(stm, "                     eg.\n");
    fprintf(stm, "                       -r\"/usr/include:/etc\"\n");
#else
# error Unknown platform
#endif /* platform */
    fprintf(stm, "    -h             - searches from the roots of all drives on the system\n");
    fprintf(stm, "    -R             - suppresses recursive search\n");
    fprintf(stm, "    -s             - succinct output. Prints path only\n");
    fprintf(stm, "    -u             - recursive search. (Default except for environment variable searches.)\n");
    fprintf(stm, "    -v             - verbose output. Prints time, attributes, size and path. (default)\n");
    fprintf(stm, "    -w             - searches from the current working directory\n");
    fprintf(stm, "    <search-spec>  - one or more file search specifications, separated by \';\',\n");
    fprintf(stm, "                     eg.\n");
    fprintf(stm, "                       \"*.exe\"\n");
    fprintf(stm, "                       \"myfile.ext\"\n");
    fprintf(stm, "                       \"*.exe;*.dll\"\n");
    fprintf(stm, "                       \"*.xl?;report.*\"\n");
    fprintf(stm, "\n");
    fprintf(stm, "  USAGE 2: %s -?\n", EXE_NAME);
    fprintf(stm, "\n");
    fprintf(stm, "    Displays this help\n");
    fprintf(stm, "\n");
    fprintf(stm, "\n");
    fprintf(stm, "  Contact %s\n", _nameSynesisSoftware);
    fprintf(stm, "    at \"%s\",\n", _wwwSynesisSoftware_SystemTools);
    fprintf(stm, "    or, via email, at \"%s\"\n", _emailSynesisSoftware_SystemTools);
    fprintf(stm, "\n");
    fprintf(stm, "  Contact Digital Mars\n");
    fprintf(stm, "    at \"www.digitalmars.com\",\n");
    fprintf(stm, "    or, via email, at \"software@digitalmars.com\"\n");
    fprintf(stm, "\n");

    if(bExit)
    {
        exit(EXIT_FAILURE);
    }
}

#ifdef WHEREIS_PLATFORM_IS_WIN32
char const *get_driveroots(char *const buffer)
{
    char    drv[4] = "?:\\";

    *buffer = '\0';

    for(char ch = 'a'; ch <= 'z'; ++ch)
    {
        drv[0] = ch;

        long    type    =   static_cast<long>(GetDriveTypeA(drv));

        if(DRIVE_FIXED == type)
        {
            strcat(buffer, drv);
            strcat(buffer, ";");
        }
    }

    return buffer;
}
#endif /* WHEREIS_PLATFORM_IS_WIN32 */

/* ////////////////////////////////////////////////////////////////////////// */
