/* /////////////////////////////////////////////////////////////////////////////
 * File:        array_view_test.cpp
 *
 * Purpose:     Implementation file for the array_view_test project.
 *
 * Created:     28th October 2005
 * Updated:     28th October 2005
 *
 * Status:      Wizard-generated
 *
 * License:     (Licensed under the Synesis Software Open License)
 *
 *              Copyright 1999-2005, Synesis Software Pty Ltd.
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/software
 *
 *              email:      software@synesis.com.au
 *
 *              This source code is placed into the public domain 2005
 *              by Synesis Software Pty Ltd. There are no restrictions
 *              whatsoever to your use of the software. 
 *
 *              This source code is provided by Synesis Software Pty Ltd "as is"
 *              and any warranties, whether expressed or implied, including, but
 *              not limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software Pty Ltd be liable for any direct,
 *              indirect, incidental, special, exemplary, or consequential
 *              damages (including, but not limited to, procurement of
 *              substitute goods or services; loss of use, data, or profits; or
 *              business interruption) however caused and on any theory of
 *              liability, whether in contract, strict liability, or tort
 *              (including negligence or otherwise) arising in any way out of
 *              the use of this software, even if advised of the possibility of
 *              such damage. 
 *
 *              Neither the name of Synesis Software Pty Ltd nor the names of
 *              any subdivisions, employees or agents of Synesis Software Pty
 *              Ltd, nor the names of any other contributors to this software
 *              may be used to endorse or promote products derived from this
 *              software without specific prior written permission. 
 *
 * ////////////////////////////////////////////////////////////////////////// */

/* STLSoft Header Files */
#include <stlsoft.h>
#include <stlsoft/array_view.hpp>
#include <stlsoft/integer_to_string.hpp>

/* Standard C++ Header Files */
#include <algorithm>
#include <exception>
#include <iostream>
#include <string>
#include <vector>
#if 0
#include <iterator>
#include <list>
#endif /* 0 */

#if !defined(__WATCOMC__) && \
    (   !defined(_MSC_VER) || \
        _MSC_VER >= 1100)

using std::cerr;
using std::cin;
using std::cout;
using std::endl;

#else /* ? __WATCOMC__ */
namespace std
{
    using ::exception;
}
#endif /* __WATCOMC__ */


/* Standard C Header Files */
#include <stdlib.h>
#include <time.h>

#if defined(_MSC_VER) && \
	defined(_DEBUG)
# include <crtdbg.h>
#endif /* _MSC_VER) && _DEBUG */

/* /////////////////////////////////////////////////////////////////////////////
 * Typedefs
 */

typedef std::string						string_t;

typedef std::vector<string_t>			strings_t;
typedef stlsoft::array_view<string_t>	strings_view_t;

typedef std::vector<int>				ints_t;
typedef stlsoft::array_view<int>		ints_view_t;

typedef std::vector<double>				doubles_t;
typedef stlsoft::array_view<double>		doubles_view_t;

/* /////////////////////////////////////////////////////////////////////////////
 * Forward declarations
 */

template <typename C>
void dump_collection(char const *message, C const &coll)
{
	C	coll2(coll);
	C	coll3;

	coll2.swap(coll3);

	cout << "Dumping: " << message << endl << endl;

	cout << "1. via for_each(begin(), end())" << endl;
	{ for(ss_typename_type_k C::const_iterator b = coll.begin(); b != coll.end(); ++b)
	{
		cout << "  " << *b << endl;
	}}
	cout << endl;

	cout << "1. via for_each(rbegin(), rend())" << endl;
	{ for(ss_typename_type_k C::const_reverse_iterator b = coll.rbegin(); b != coll.rend(); ++b)
	{
		cout << "  " << *b << endl;
	}}
	cout << endl;

	cout << "2. via subscript operator" << endl;
	{ for(size_t i = 0; i < coll.size(); ++i)
	{
		cout << "  " << coll[i] << endl;
	}}
	cout << endl;
}

/* ////////////////////////////////////////////////////////////////////////// */

static int main_(int /* argc */, char ** /*argv*/)
{
	/* . */

	strings_t		strings(8);
	strings_view_t	strings_view(&strings[0], strings.size());

	{ for(strings_t::iterator it = strings.begin(); it != strings.end(); ++it)
	{
		char	sz[21];

		*it = stlsoft::integer_to_string(&sz[0], stlsoft_num_elements(sz), ::time(NULL) & 0xff);
	}}

	dump_collection("std::vector<string>", strings);
	dump_collection("stlsoft::array_view<string>", strings_view);

	ints_t			ints(10);
	ints_view_t		ints_view(&ints[0], ints.size());
	
	std::generate(ints.begin(), ints.end(), ::rand);

	dump_collection("std::vector<int>", ints);
	dump_collection("stlsoft::array_view<int>", ints_view);

	doubles_t		doubles(5);
	doubles_view_t	doubles_view(&doubles[0], doubles.size());

	dump_collection("std::vector<double>", doubles);
	dump_collection("stlsoft::array_view<double>", doubles_view);

    return EXIT_SUCCESS;
}

int main(int argc, char *argv[])
{
	int				iRet;

#if defined(_MSC_VER) && \
    defined(_DEBUG)
    _CrtMemState	memState;
#endif /* _MSC_VER && _MSC_VER */

#if defined(_MSC_VER) && \
    defined(_DEBUG)
    _CrtMemCheckpoint(&memState);
#endif /* _MSC_VER && _MSC_VER */

#if 0
	{ for(size_t i = 0; i < 0xffffffff; ++i){} }
#endif /* 0 */

	try
	{
#if defined(_DEBUG) || \
    defined(__SYNSOFT_DBS_DEBUG)
		cout << "array_view_test: " << __STLSOFT_COMPILER_LABEL_STRING << endl;
#endif /* debug */


		iRet = main_(argc, argv);
	}
	catch(std::exception &x)
	{
		cerr << "Unhandled error: " << x.what() << endl;

		iRet = EXIT_FAILURE;
	}
	catch(...)
	{
		cerr << "Unhandled unknown error" << endl;

		iRet = EXIT_FAILURE;
	}

#if defined(_MSC_VER) && \
	defined(_DEBUG)
    _CrtMemDumpAllObjectsSince(&memState);
#endif /* _MSC_VER) && _DEBUG */

	return iRet;
}

/* ////////////////////////////////////////////////////////////////////////// */
